import { Subscribable, Observable, Subject, Observer, Unsubscribable } from 'rxjs';
import { CommandResult } from './command_result.js';
import { CommandError } from './command_error.js';
import { Action, AsyncAction, ObservableAction, RxCommandOptions } from './types.js';

/**
 * [RxCommand] capsules a given handler function that can then be executed by its [execute] method.
 * The result of this method is then published through its Observable (Observable wrap Dart Streams)
 * Additionally it offers Observables for it's current execution state, if the command can be executed and for
 * all possibly thrown exceptions during command execution.
 *
 * [RxCommand] implements the `Observable` interface so you can listen directly to the [RxCommand] which emits the
 * results of the wrapped function. If this function has a [void] return type
 * it will still output one `void` item so that you can listen for the end of the execution.
 *
 * The [results] Observable emits [CommandResult<TRESULT>] which is often easier in combination with Flutter `StreamBuilder`
 * because you have all state information at one place.
 *
 * An [RxCommand] is a generic class of type [RxCommand<TParam, TRESULT>]
 * where [TParam] is the type of data that is passed when calling [execute] and
 * [TResult] denotes the return type of the handler function. To signal that
 * a handler doesn't take a parameter or returns no value use the type `void`
 */
declare class RxCommand<TParam = void, TResult = void> extends Function implements Subscribable<TResult> {
    private readonly _action;
    private readonly _resultsSubject;
    private readonly _restriction;
    private readonly _includeLastResultInCommandResults;
    private readonly _resultSubjectIsBehaviourSubject;
    private readonly _emitInitialCommandResult;
    private readonly _debugName;
    /**
     * The result of the last successful call to execute. This is especially handy to use as `initialData` of Flutter `StreamBuilder`
     * */
    lastResult?: TResult;
    private _running;
    private _canExecut;
    private _executionLocked;
    private readonly _commandResultsSubject;
    private readonly _isExecutingSubject;
    private readonly _canExecuteSubject;
    private readonly _thrownExceptionsSubject;
    private readonly _subscriptions;
    private set _sink(value);
    get observable(): Observable<TResult>;
    get results(): Observable<CommandResult<TParam, TResult>>;
    get isExecuting(): Observable<boolean>;
    get canExecute(): Observable<boolean>;
    get thrownExceptions(): Observable<CommandError<TParam>>;
    get next(): Promise<TResult | CommandError<TParam>>;
    /**
     * optional hander that will get called on any exception that happens inside
     * any Command of the app. Ideal for logging.
     */
    static globalExceptionHandler?: (error?: CommandError<unknown>) => void;
    /**
     * optional handler that will get called on all `Command` executions.
     */
    static loggingHandler?: (result?: CommandResult<unknown, unknown>) => void;
    constructor(_action: Action<TParam, TResult> | AsyncAction<TParam, TResult> | ObservableAction<TParam, TResult>, _resultsSubject: Subject<TResult>, _restriction: Observable<boolean> | undefined, _includeLastResultInCommandResults: boolean, _resultSubjectIsBehaviourSubject: boolean, _emitInitialCommandResult: boolean, _debugName: string | undefined, lastResult?: TResult);
    /**
     * Creates a RxCommand from a sync function, a async function or an "one time" observable.
     * [provider]: provider function that returns a value, a Promise or a Observable that will be subscribed on the call of [execute]
     * [restriction] : observable that can be used to enable/disable the command based on some other state change
     * if omitted the command can be executed always except it's already executing
     * [isExecuting] will issue a `bool` value on each state change. Even if you
     * subscribe to a newly created command it will issue `false`
     * For the `Observable<CommandResult>` that [RxCommand] publishes in [results] this normally doesn't make sense
     * if you want to get an initial Result with `data==null, error==null, isExecuting==false` pass
     * [emitInitialCommandResult=true].
     * [emitLastResult] will include the value of the last successful execution in all [CommandResult] events unless there is no result.
     * By default the [results] Observable and the [RxCommand] itself behave like a Subject. If you want that it acts like
     * a BehaviourSubject, meaning every listener gets the last received value, you can set [emitsLastValueToNewSubscriptions = true].
     * [initialLastResult] sets the value of the [lastResult] property before the first item was received.
     * [debugName] optional identifier that is included when you register a [globalExceptionHandler]
     * or a [loggingHandler]
     */
    static create<TParam = void, TResult = void>(action: Action<TParam, TResult> | AsyncAction<TParam, TResult> | ObservableAction<TParam, TResult>, options?: RxCommandOptions<TResult>): RxCommand<TParam, TResult>;
    /**
     * Helper for create a RxCommand from a sync function, a async function or an "one time" observable without param inputs but with return type.
     * [provider]: provider function that returns a value, a Promise or a Observable that will be subscribed on the call of [execute]
     * [restriction] : observable that can be used to enable/disable the command based on some other state change
     * if omitted the command can be executed always except it's already executing
     * [isExecuting] will issue a `bool` value on each state change. Even if you
     * subscribe to a newly created command it will issue `false`
     * For the `Observable<CommandResult>` that [RxCommand] publishes in [results] this normally doesn't make sense
     * if you want to get an initial Result with `data==null, error==null, isExecuting==false` pass
     * [emitInitialCommandResult=true].
     * [emitLastResult] will include the value of the last successful execution in all [CommandResult] events unless there is no result.
     * By default the [results] Observable and the [RxCommand] itself behave like a Subject. If you want that it acts like
     * a BehaviourSubject, meaning every listener gets the last received value, you can set [emitsLastValueToNewSubscriptions = true].
     * [initialLastResult] sets the value of the [lastResult] property before the first item was received.
     * [debugName] optional identifier that is included when you register a [globalExceptionHandler]
     * or a [loggingHandler]
     */
    static createNoParam<TResult, TParam = void>(action: Action<TParam, TResult> | AsyncAction<TParam, TResult> | ObservableAction<TParam, TResult>, options?: RxCommandOptions<TResult>): RxCommand<TParam, TResult>;
    subscribe(observer: Partial<Observer<TResult>> | ((value: TResult) => void) | undefined): Unsubscribable;
    _call(...args: TParam[]): void;
    execute(param: TParam): void;
    dispose(): void;
    private _logger;
    private _debug;
}

export { RxCommand };
