import {
  BehaviorSubject,
  distinct,
  filter,
  from,
  lastValueFrom,
  merge,
  Observable,
  of,
  ReplaySubject,
  Subject,
  Subscription,
  switchMap,
  take,
  tap
} from "rxjs";
import { match, P } from "ts-pattern";
import { CommandResult } from "./command_result.js";
import { CommandError } from "./command_error.js";
const DEFAULT_OPTIONS = {
  emitInitialCommandResult: false,
  emitLastResult: false,
  emitsLastValueToNewSubscriptions: false
};
class RxCommand extends Function {
  constructor(_action, _resultsSubject, _restriction, _includeLastResultInCommandResults, _resultSubjectIsBehaviourSubject, _emitInitialCommandResult, _debugName, lastResult) {
    super("return arguments.callee._call.apply(arguments.callee, arguments)");
    this._action = _action;
    this._resultsSubject = _resultsSubject;
    this._restriction = _restriction;
    this._includeLastResultInCommandResults = _includeLastResultInCommandResults;
    this._resultSubjectIsBehaviourSubject = _resultSubjectIsBehaviourSubject;
    this._emitInitialCommandResult = _emitInitialCommandResult;
    this._debugName = _debugName;
    this._commandResultsSubject = this._resultSubjectIsBehaviourSubject ? new ReplaySubject(1) : new Subject();
    this.lastResult = lastResult;
    this._sink = this._commandResultsSubject.pipe(filter((x) => x.hasError)).subscribe(
      (x) => this._thrownExceptionsSubject.next(
        new CommandError(x.param, x.error, this._debugName)
      )
    );
    this._sink = this._commandResultsSubject.subscribe(
      (x) => this._isExecutingSubject.next(x.isExecuting)
    );
    const _canExecuteParam = this._restriction == void 0 ? of(true) : this._restriction.pipe(
      tap({
        error: (error) => {
          if (error instanceof Error) {
            this._thrownExceptionsSubject.next(
              new CommandError(void 0, error, this._debugName)
            );
          }
        }
      }),
      distinct()
    );
    this._sink = _canExecuteParam.subscribe((canExecute) => {
      this._canExecut = canExecute && !this._running;
      this._executionLocked = !canExecute;
      this._canExecuteSubject.next(this._canExecut);
    });
    if (this._emitInitialCommandResult) {
      this._commandResultsSubject.next(
        new CommandResult(
          void 0,
          this.lastResult,
          void 0,
          false
        )
      );
    }
  }
  /**
   * The result of the last successful call to execute. This is especially handy to use as `initialData` of Flutter `StreamBuilder`
   * */
  lastResult;
  _running = false;
  _canExecut = true;
  _executionLocked = false;
  _commandResultsSubject;
  _isExecutingSubject = new BehaviorSubject(false);
  _canExecuteSubject = new BehaviorSubject(false);
  _thrownExceptionsSubject = new Subject();
  _subscriptions = new Subscription();
  set _sink(subscription) {
    this._subscriptions.add(subscription);
  }
  get observable() {
    return this._resultsSubject;
  }
  get results() {
    return this._commandResultsSubject;
  }
  get isExecuting() {
    return this._isExecutingSubject;
  }
  get canExecute() {
    return this._canExecuteSubject;
  }
  get thrownExceptions() {
    return this._thrownExceptionsSubject;
  }
  get next() {
    const observable = merge(this._resultsSubject, this.thrownExceptions).pipe(
      take(1)
    );
    return lastValueFrom(observable);
  }
  /**
   * optional hander that will get called on any exception that happens inside
   * any Command of the app. Ideal for logging.
   */
  static globalExceptionHandler;
  /**
   * optional handler that will get called on all `Command` executions.
   */
  static loggingHandler;
  /**
   * Creates a RxCommand from a sync function, a async function or an "one time" observable.
   * [provider]: provider function that returns a value, a Promise or a Observable that will be subscribed on the call of [execute]
   * [restriction] : observable that can be used to enable/disable the command based on some other state change
   * if omitted the command can be executed always except it's already executing
   * [isExecuting] will issue a `bool` value on each state change. Even if you
   * subscribe to a newly created command it will issue `false`
   * For the `Observable<CommandResult>` that [RxCommand] publishes in [results] this normally doesn't make sense
   * if you want to get an initial Result with `data==null, error==null, isExecuting==false` pass
   * [emitInitialCommandResult=true].
   * [emitLastResult] will include the value of the last successful execution in all [CommandResult] events unless there is no result.
   * By default the [results] Observable and the [RxCommand] itself behave like a Subject. If you want that it acts like
   * a BehaviourSubject, meaning every listener gets the last received value, you can set [emitsLastValueToNewSubscriptions = true].
   * [initialLastResult] sets the value of the [lastResult] property before the first item was received.
   * [debugName] optional identifier that is included when you register a [globalExceptionHandler]
   * or a [loggingHandler]
   */
  static create(action, options) {
    const config = { ...DEFAULT_OPTIONS, ...options };
    const subject = config.emitsLastValueToNewSubscriptions || config.emitInitialCommandResult ? new ReplaySubject(1) : new Subject();
    return new RxCommand(
      action,
      subject,
      config.restriction,
      config.emitLastResult,
      config.emitsLastValueToNewSubscriptions || config.emitInitialCommandResult,
      config.emitInitialCommandResult,
      config.debugName,
      config.initialLastResult
    );
  }
  /**
   * Helper for create a RxCommand from a sync function, a async function or an "one time" observable without param inputs but with return type.
   * [provider]: provider function that returns a value, a Promise or a Observable that will be subscribed on the call of [execute]
   * [restriction] : observable that can be used to enable/disable the command based on some other state change
   * if omitted the command can be executed always except it's already executing
   * [isExecuting] will issue a `bool` value on each state change. Even if you
   * subscribe to a newly created command it will issue `false`
   * For the `Observable<CommandResult>` that [RxCommand] publishes in [results] this normally doesn't make sense
   * if you want to get an initial Result with `data==null, error==null, isExecuting==false` pass
   * [emitInitialCommandResult=true].
   * [emitLastResult] will include the value of the last successful execution in all [CommandResult] events unless there is no result.
   * By default the [results] Observable and the [RxCommand] itself behave like a Subject. If you want that it acts like
   * a BehaviourSubject, meaning every listener gets the last received value, you can set [emitsLastValueToNewSubscriptions = true].
   * [initialLastResult] sets the value of the [lastResult] property before the first item was received.
   * [debugName] optional identifier that is included when you register a [globalExceptionHandler]
   * or a [loggingHandler]
   */
  static createNoParam(action, options) {
    return RxCommand.create(action, options);
  }
  subscribe(observer) {
    return this._resultsSubject.subscribe(observer);
  }
  _call(...args) {
    return this.execute(args[0]);
  }
  execute(param) {
    if (!this._canExecut || this._running) {
      return;
    }
    this._running = true;
    this._canExecuteSubject.next(false);
    this._commandResultsSubject.next(
      new CommandResult(
        param,
        this._includeLastResultInCommandResults ? this.lastResult : void 0,
        void 0,
        true
      )
    );
    this._sink = of(true).pipe(
      switchMap(() => {
        const data = this._action(param);
        return match(data).returnType().with(P.instanceOf(Promise), () => from(data)).with(P.instanceOf(Observable), () => data).otherwise(() => of(data));
      })
    ).subscribe({
      next: (data) => {
        const result = new CommandResult(
          param,
          data,
          void 0,
          false
        );
        this._resultsSubject.next(data);
        this._commandResultsSubject.next(result);
        this._logger(result);
      },
      error: (error) => {
        if (error instanceof Error) {
          const result = new CommandResult(
            param,
            this._includeLastResultInCommandResults ? this.lastResult : void 0,
            error,
            false
          );
          this._commandResultsSubject.next(result);
          this._debug(param, error);
          this._logger(result);
        }
      },
      complete: () => {
        this._running = false;
        this._canExecut = !this._executionLocked;
        this._canExecuteSubject.next(!this._executionLocked);
      }
    });
  }
  dispose() {
    this._commandResultsSubject.complete();
    this._isExecutingSubject.complete();
    this._canExecuteSubject.complete();
    this._thrownExceptionsSubject.complete();
    this._resultsSubject.complete();
    this._subscriptions.unsubscribe();
  }
  _logger(result) {
    if (RxCommand.loggingHandler != void 0) {
      RxCommand.loggingHandler(result);
    }
  }
  _debug(param, error) {
    if (RxCommand.globalExceptionHandler != null) {
      RxCommand.globalExceptionHandler(
        new CommandError(param, error, this._debugName)
      );
    }
  }
}
export {
  RxCommand
};
